#include <avr/io.h>
#include <stdio.h>
#include <util/delay.h>
#include <avr/interrupt.h>

/************************************************************************/
/*                          macro-urile generale                        */
/************************************************************************/
#define WRITE_ONE(REG, PIN)		REG |= (1 << PIN);
#define WRITE_ZERO(REG, PIN)	REG &= ~(1 << PIN);
#define TOGGLE_BIT(REG, PIN)	REG ^= (1 << PIN);

#define READ_BIT(REG_R, PIN)	(REG_R & (1 << PIN))

#define INPUT_PIN(DDR, PIN)		WRITE_ZERO(DDR, PIN)
#define OUTPUT_PIN(DDR, PIN)	WRITE_ONE(DDR, PIN)

#define REG_LED					PORTD
#define PIN_LED					PIND7


#define PIN_ECHO		PINA7
#define PIN_TRIG		PINA6
#define REG_HC_SR04		PORTA
#define DELAY_HC_SR04	10
/** viteza in cm / mcrosecunde */
#define SOUND_SPEED		0.0343f
#define MAX_DIST		10.0f
#define MIN_DIST		35.0f

static volatile int number_overflow = 0;
static volatile int time_read = 0;
static volatile int finsh_read = 0;
static volatile int start_measure = 0;

#define IN1				PIND0
#define IN2				PIND1
#define IN3				PIND2
#define	IN4				PIND3

#define EN0				PIND4
#define EN1				PIND5

#define LEFT			0
#define RIGHT			1

#define PIN_BUZZER		PINB6
#define REG_BUZZER		PORTB

ISR(TIMER1_OVF_vect)
{
	number_overflow++;
}

ISR(PCINT0_vect) 
{
	
	if (READ_BIT(PINA, PIN_ECHO) != 0) {
		if (start_measure == 0) {
			start_measure = 1;
		}
	} else if (!READ_BIT(PINA, PIN_ECHO) && start_measure && !finsh_read) {
		time_read = TCNT1;
		finsh_read = 1;
	}
}

void set_time_hc_sr04(void) {
	/** am ales timer-ul 1 pentru ca are o rezolutie mai mare */
	 // activeaza intreruperea de Overflow.
	TIMSK1 = (1 << TOIE1);
	/* prescaler-ul de 8, count incrementat la  0.5uS */
	TCCR1B = (1 << CS11);
	
	// intreuperea va fi pe pinul PINA7
	PCICR |= (1 << PCIE0);
	PCIFR |= (1 << PCIF0);
	PCMSK0 |= (1 << PCINT7);
	
	sei();
}

float get_distance() {
	float time_sem = 1.0f;
		
	finsh_read = 0;
	start_measure = 0;
	number_overflow = 0;
	TCNT1 = 0;
	
	WRITE_ONE(REG_HC_SR04, PIN_TRIG);
	_delay_us(DELAY_HC_SR04);
	WRITE_ZERO(REG_HC_SR04, PIN_TRIG);
	
	while (!finsh_read);
	
	time_sem = 1.0f * (time_read + number_overflow * 0xFFF) / 2;
	
	finsh_read = 0;
	start_measure = 0;
	number_overflow = 0;
	TCNT1 = 0;
	
	return time_sem  * SOUND_SPEED / 2.0f;
}

void set_pin(void) {
	/* pt led */
	OUTPUT_PIN(DDRD, PIN_LED);
	WRITE_ZERO(REG_LED, PIN_LED);
	
	/* pt HC-SR04 */
	OUTPUT_PIN(REG_HC_SR04, PIN_TRIG);
	WRITE_ZERO(REG_HC_SR04, PIN_TRIG);
	INPUT_PIN(REG_HC_SR04, PIN_ECHO);
	
	/** pt L298N */
	OUTPUT_PIN(DDRD, IN1);
	OUTPUT_PIN(DDRD, IN2);
	OUTPUT_PIN(DDRD, IN3);
	OUTPUT_PIN(DDRD, IN4);
	OUTPUT_PIN(DDRD, EN0);
	OUTPUT_PIN(DDRD, EN1);
	
	WRITE_ZERO(PORTD, IN1);
	WRITE_ZERO(PORTD, IN2);
	WRITE_ZERO(PORTD, IN3);
	WRITE_ZERO(PORTD, IN4);
	WRITE_ZERO(PORTD, EN0);
	WRITE_ZERO(PORTD, EN1);
	
	/** pt buzzer */
	OUTPUT_PIN(DDRB, PIN_BUZZER);
	WRITE_ONE(REG_BUZZER, PIN_BUZZER);
}

void signal_led(void) {
	WRITE_ONE(REG_LED, PIN_LED);
	WRITE_ZERO(REG_BUZZER, PIN_BUZZER);
	_delay_ms(1000);
	WRITE_ZERO(REG_LED, PIN_LED);
	WRITE_ONE(REG_BUZZER, PIN_BUZZER);
	_delay_ms(500);
}

void reset_motor(void) {
	_delay_ms(150);
	WRITE_ZERO(PORTD, IN1);
	WRITE_ZERO(PORTD, IN2);
	WRITE_ZERO(PORTD, IN3);
	WRITE_ZERO(PORTD, IN4);
	_delay_ms(50);
}

void forward() {
	/** LEFT */
	WRITE_ONE(PORTD, IN1);
	WRITE_ZERO(PORTD, IN2);

	/** RIGHT */
	WRITE_ONE(PORTD, IN3);
	WRITE_ZERO(PORTD, IN4);
	reset_motor();
}

void turn_left(void) {
	/** RIGHT forward */
	WRITE_ONE(PORTD, IN3);
	WRITE_ZERO(PORTD, IN4);
	
	/** back LEFT */
	WRITE_ONE(PORTD, IN2);
	WRITE_ZERO(PORTD, IN1);
	reset_motor();
}


int main()
{	
	set_pin();
	set_time_hc_sr04();
	_delay_ms(1000);
	
	while(1) {
		float dist = get_distance();
		
		if (dist < MAX_DIST) {
			signal_led();
			float new_dist = 0;
			
			while (new_dist < MIN_DIST) {
				turn_left();
				new_dist = get_distance();
			}
		}
		
		/** deplasarea in fata */
		forward();
	}
	return 0;
}